from deepmerge.strategy.set import SetStrategies


def test_union_unions():
    assert SetStrategies.strategy_union({}, [], set("abc"), set("bcd")) == set("abcd")


def test_intersect_intersects():
    assert SetStrategies.strategy_intersect({}, [], set("abc"), set("bcd")) == set("bc")


def test_override_overrides():
    assert SetStrategies.strategy_override({}, [], set("abc"), set("bcd")) == set("bcd")



from .core import StrategyList


class ListStrategies(StrategyList):
    """
    Contains the strategies provided for lists.
    """

    NAME = "list"

    @staticmethod
    def strategy_override(config, path, base, nxt):
        """use the list nxt."""
        return nxt

    @staticmethod
    def strategy_prepend(config, path, base, nxt):
        """prepend nxt to base."""
        return nxt + base

    @staticmethod
    def strategy_append(config, path, base, nxt):
        """append nxt to base."""
        return base + nxt

    @staticmethod
    def strategy_append_unique(config, path, base, nxt):
        """append items without duplicates in nxt to base."""
        base_as_set = set(base)
        return base + [n for n in nxt if n not in base_as_set]



from deepmerge.strategy.type_conflict import TypeConflictStrategies

EMPTY_DICT = {}

CONTENT_AS_LIST = [{"key": "val"}]


def test_merge_if_not_empty():
    strategy = TypeConflictStrategies.strategy_override_if_not_empty(
        {}, [], EMPTY_DICT, CONTENT_AS_LIST
    )
    assert strategy == CONTENT_AS_LIST

    strategy = TypeConflictStrategies.strategy_override_if_not_empty(
        {}, [], CONTENT_AS_LIST, EMPTY_DICT
    )
    assert strategy == CONTENT_AS_LIST

    strategy = TypeConflictStrategies.strategy_override_if_not_empty(
        {}, [], CONTENT_AS_LIST, None
    )
    assert strategy == CONTENT_AS_LIST



import pytest
from deepmerge import Merger


@pytest.fixture
def custom_merger():
    def merge_sets(merger, path, base, nxt):
        base |= nxt
        return base

    def merge_list(merger, path, base, nxt):
        if len(nxt) > 0:
            base.append(nxt[-1])
            return base

    return Merger(
        [(list, merge_list), (dict, "merge"), (set, merge_sets)],
        [],
        [],
    )


def test_custom_merger_applied(custom_merger):
    result = custom_merger.merge({"foo"}, {"bar"})
    assert result == {"foo", "bar"}


def test_custom_merger_list(custom_merger):
    result = custom_merger.merge([1, 2, 3], [4, 5, 6])
    assert result == [1, 2, 3, 6]



from .core import StrategyList


class SetStrategies(StrategyList):
    """
    Contains the strategies provided for sets.
    """

    NAME = "set"

    @staticmethod
    def strategy_union(config, path, base, nxt):
        """
        use all values in either base or nxt.
        """
        return base | nxt

    @staticmethod
    def strategy_intersect(config, path, base, nxt):
        """
        use all values in both base and nxt.
        """
        return base & nxt

    @staticmethod
    def strategy_override(config, path, base, nxt):
        """
        use the set nxt.
        """
        return nxt



from .core import StrategyList


class DictStrategies(StrategyList):
    """
    Contains the strategies provided for dictionaries.

    """

    NAME = "dict"

    @staticmethod
    def strategy_merge(config, path, base, nxt):
        """
        for keys that do not exists,
        use them directly. if the key exists
        in both dictionaries, attempt a value merge.
        """
        for k, v in nxt.items():
            if k not in base:
                base[k] = v
            else:
                base[k] = config.value_strategy(path + [k], base[k], v)
        return base

    @staticmethod
    def strategy_override(config, path, base, nxt):
        """
        move all keys in nxt into base, overriding
        conflicts.
        """
        return nxt



from ..exception import StrategyNotFound, InvalidMerge
from ..compat import string_type

STRATEGY_END = object()


class StrategyList(object):

    NAME = None

    def __init__(self, strategy_list):
        if not isinstance(strategy_list, list):
            strategy_list = [strategy_list]
        self._strategies = [self._expand_strategy(s) for s in strategy_list]

    @classmethod
    def _expand_strategy(cls, strategy):
        """
        :param strategy: string or function

        If the strategy is a string, attempt to resolve it
        among the built in strategies.

        Otherwise, return the value, implicitly assuming it's a function.
        """
        if isinstance(strategy, string_type):
            method_name = "strategy_{0}".format(strategy)
            if not hasattr(cls, method_name):
                raise StrategyNotFound(strategy)
            return getattr(cls, method_name)
        return strategy

    def __call__(self, *args, **kwargs):
        for s in self._strategies:
            ret_val = s(*args, **kwargs)
            if ret_val is not STRATEGY_END:
                return ret_val
        raise InvalidMerge(self.NAME, args, kwargs)



from .core import StrategyList


class FallbackStrategies(StrategyList):
    """
    The StrategyList containing fallback strategies.
    """

    NAME = "fallback"

    @staticmethod
    def strategy_override(config, path, base, nxt):
        """use nxt, and ignore base."""
        return nxt

    @staticmethod
    def strategy_use_existing(config, path, base, nxt):
        """use base, and ignore next."""
        return base



from .merger import Merger
from .strategy.core import STRATEGY_END  # noqa

# some standard mergers available

DEFAULT_TYPE_SPECIFIC_MERGE_STRATEGIES = [
    (list, "append"),
    (dict, "merge"),
    (set, "union"),
]

# this merge will never raise an exception.
# in the case of type mismatches,
# the value from the second object
# will override the previous one.
always_merger = Merger(
    DEFAULT_TYPE_SPECIFIC_MERGE_STRATEGIES, ["override"], ["override"]
)

# this merge strategies attempts
# to merge (append for list, unify for dicts)
# if possible, but raises an exception
# in the case of type conflicts.
merge_or_raise = Merger(DEFAULT_TYPE_SPECIFIC_MERGE_STRATEGIES, [], [])

# a conservative merge tactic:
# for data structures with a specific
# strategy, keep the existing value.
# similar to always_merger but instead
# keeps existing values when faced
# with a type conflict.
conservative_merger = Merger(
    DEFAULT_TYPE_SPECIFIC_MERGE_STRATEGIES, ["use_existing"], ["use_existing"]
)



import pytest
from deepmerge.strategy.list import ListStrategies
from deepmerge import Merger


@pytest.fixture
def custom_merger():
    return Merger(
        [(list, ListStrategies.strategy_append_unique)],
        [],
        [],
    )


def test_strategy_append_unique(custom_merger):
    base = [1, 3, 2]
    nxt = [3, 5, 4, 1, 2]

    expected = [1, 3, 2, 5, 4]
    actual = custom_merger.merge(base, nxt)
    assert actual == expected



from .strategy.list import ListStrategies
from .strategy.dict import DictStrategies
from .strategy.set import SetStrategies
from .strategy.type_conflict import TypeConflictStrategies
from .strategy.fallback import FallbackStrategies


class Merger(object):
    """
    :param type_strategies, List[Tuple]: a list of (Type, Strategy) pairs
           that should be used against incoming types. For example: (dict, "override").
    """

    PROVIDED_TYPE_STRATEGIES = {
        list: ListStrategies,
        dict: DictStrategies,
        set: SetStrategies,
    }

    def __init__(self, type_strategies, fallback_strategies, type_conflict_strategies):
        self._fallback_strategy = FallbackStrategies(fallback_strategies)

        expanded_type_strategies = []
        for typ, strategy in type_strategies:
            if typ in self.PROVIDED_TYPE_STRATEGIES:
                strategy = self.PROVIDED_TYPE_STRATEGIES[typ](strategy)
            expanded_type_strategies.append((typ, strategy))
        self._type_strategies = expanded_type_strategies

        self._type_conflict_strategy = TypeConflictStrategies(type_conflict_strategies)

    def merge(self, base, nxt):
        return self.value_strategy([], base, nxt)

    def type_conflict_strategy(self, *args):
        return self._type_conflict_strategy(self, *args)

    def value_strategy(self, path, base, nxt):
        if not (isinstance(base, type(nxt)) or isinstance(nxt, type(base))):
            return self.type_conflict_strategy(path, base, nxt)
        for typ, strategy in self._type_strategies:
            if isinstance(nxt, typ):
                return strategy(self, path, base, nxt)
        return self._fallback_strategy(self, path, base, nxt)



from .core import StrategyList


class TypeConflictStrategies(StrategyList):
    """contains the strategies provided for type conflicts."""

    NAME = "type conflict"

    @staticmethod
    def strategy_override(config, path, base, nxt):
        """overrides the new object over the old object"""
        return nxt

    @staticmethod
    def strategy_use_existing(config, path, base, nxt):
        """uses the old object instead of the new object"""
        return base

    @staticmethod
    def strategy_override_if_not_empty(config, path, base, nxt):
        """overrides the new object over the old object only if the new object is not empty or null"""
        return nxt if nxt else base



from deepmerge.strategy.core import StrategyList
from deepmerge import STRATEGY_END


def return_true_if_foo(config, path, base, nxt):
    if base == "foo":
        return True
    return STRATEGY_END


def always_return_custom(config, path, base, nxt):
    return "custom"


def test_single_value_allowed():
    """ """

    def strat(name):
        return name

    sl = StrategyList(strat)
    assert sl("foo") == "foo"


def test_first_working_strategy_is_used():
    """
    In the case where the StrategyList has multiple values,
    the first strategy which returns a valid value (i.e. not STRATEGY_END)
    should be returned.
    """
    sl = StrategyList(
        [
            return_true_if_foo,
            always_return_custom,
        ]
    )
    # return_true_if_foo will take.
    assert sl({}, [], "foo", "bar") is True
    # return_true_if_foo will fail,
    # which will then activea always_return_custom
    assert sl({}, [], "bar", "baz") == "custom"



class DeepMergeException(Exception):
    pass


class StrategyNotFound(DeepMergeException):
    pass


class InvalidMerge(DeepMergeException):
    def __init__(self, strategy_list_name, merge_args, merge_kwargs):
        super(InvalidMerge, self).__init__(
            "no more strategies found for {0} and arguments {1}, {2}".format(
                strategy_list_name, merge_args, merge_kwargs
            )
        )
        self.strategy_list_name = strategy_list_name
        self.merge_args = merge_args
        self.merge_kwargs = merge_kwargs



from deepmerge.exception import *
import pytest


from deepmerge import (
    always_merger,
    conservative_merger,
    merge_or_raise,
)


def test_fill_missing_value():
    base = {"foo": 0, "baz": 2}
    nxt = {"bar": 1}
    always_merger.merge(base, nxt)
    assert base == {"foo": 0, "bar": 1, "baz": 2}


def test_handles_set_values_via_union():
    base = {"a": set("123"), "b": 3}
    nxt = {"a": set("2345"), "c": 1}
    always_merger.merge(base, nxt)
    assert base == {"a": set("12345"), "b": 3, "c": 1}


def test_merge_or_raise_raises_exception():
    base = {"foo": 0, "baz": 2}
    nxt = {"bar": 1, "foo": "a string!"}
    with pytest.raises(InvalidMerge) as exc_info:
        merge_or_raise.merge(base, nxt)
    exc = exc_info.value
    assert exc.strategy_list_name == "type conflict"
    assert exc.merge_args == (merge_or_raise, ["foo"], 0, "a string!")
    assert exc.merge_kwargs == {}


@pytest.mark.parametrize(
    "base, nxt, expected", [("dooby", "fooby", "dooby"), (-10, "goo", -10)]
)
def test_use_existing(base, nxt, expected):
    assert conservative_merger.merge(base, nxt) == expected


def test_example():
    base = {"foo": "value", "baz": ["a"]}
    next = {"bar": "value2", "baz": ["b"]}

    always_merger.merge(base, next)

    assert base == {"foo": "value", "bar": "value2", "baz": ["a", "b"]}
